/*
 * Author:  Terry Weissman
 *          weissman@sgi.com
 */

#define MAIN


#include "seahaven.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xutil.h>


void Punt(const char *str) {
    fprintf(stderr, "%s: %s\n", progname, str);
    exit(1);
}

static void Usage() {
    fprintf(stderr, "Usage: %s [-d display] [-speed n (default 30)]\n",
	    progname);
    exit(1);
}

static void Quit() {
    // XDestroyWindow(dpy, toplevel);
    // XFlush(dpy);
    exit(0);
}

main(int argc, char **argv) {
    const char *displayname = NULL;
    const char *name = NULL;
    XClassHint *classhint;
    XWMHints *wmhints;
    XSizeHints *sizehints;
    bool sync = false;
    int i;
    Atom wm_protocols, wm_delete_window;

    progname = argv[0];

    speedup = 30;

    for (i=1 ; i<argc ; i++) {
	if (i < argc - 1 && (strcmp(argv[i], "-d") == 0 ||
			     strcmp(argv[i], "-display") == 0)) {
	    displayname = argv[++i];
	} else if (strcmp(argv[i], "-sync") == 0) {
	    sync = true;
	} else if (strcmp(argv[i], "-speed") == 0) {
	    speedup = atoi(argv[++i]);
	    if (speedup <= 0) Usage();
	} else if (strcmp(argv[i], "-name") == 0) {
	    if (argv[++i] == NULL)
		Usage();
	    name = argv[i];
	} else {
	    Usage();
	}
    }

    score = NULL;

    dpy = XOpenDisplay(displayname);
    if (!dpy) {
	Punt("Can't open display.");
    }
    screen = DefaultScreen(dpy);
    cmap = DefaultColormap(dpy, screen);
    XSynchronize(dpy, sync);

    Visual *visual = DefaultVisual(dpy, screen);

    hascolor = (visual->c_class != StaticGray &&
		visual->c_class != GrayScale && visual->bits_per_rgb > 1);

    backpixel = GetColor("forestgreen", WhitePixel(dpy, screen));

    XSetWindowAttributes attributes;
    long valuemask = CWEventMask | CWBackPixel;
    attributes.event_mask = KeyPressMask; /* ClientMessage is always there */
    attributes.background_pixel = backpixel;

    toplevel = XCreateWindow(dpy, DefaultRootWindow(dpy), 0, 0,
			     GAMEWIDTH, GAMEHEIGHT, 1, (int) CopyFromParent,
			     InputOutput, (Visual *) CopyFromParent,
			     valuemask, &attributes);
    XSetIconName(dpy, toplevel, "Seahaven");
    XStoreName(dpy, toplevel, "Seahaven Towers");
    classhint = XAllocClassHint();
    if (!classhint)
	Punt("can't allocate memory for class hint");
    if (name)
        classhint->res_name = (char*)name;
    else {
        classhint->res_name = (char*)strrchr(progname, '/');
	if (!classhint->res_name)
	    classhint->res_name = (char*)progname;
	else classhint->res_name++;
    }
    classhint->res_class = (char*)"Seahaven";
    XSetClassHint(dpy, toplevel, classhint);
    XFree(classhint);

    /* We don't do focus management -- must tell WM to do it for us, or
       otherwise we might not get the focus when initially mapped. */
    wmhints = XAllocWMHints();
    if (!wmhints)
	Punt("can't allocate memory for window manager hints");
    wmhints->flags = InputHint | StateHint;
    wmhints->input = True;
    wmhints->initial_state = NormalState;
    XSetWMHints(dpy, toplevel, wmhints);
    XFree(wmhints);

    /* Let the WM know that we can't be resized. */
    sizehints = XAllocSizeHints();
    if (!sizehints)
	Punt("can't allocate memory for window size hints");
    sizehints->flags = PMinSize | PMaxSize;
    sizehints->min_width = sizehints->max_width = GAMEWIDTH;
    sizehints->min_height = sizehints->max_height = GAMEHEIGHT;
    XSetWMNormalHints(dpy, toplevel, sizehints);
    XFree(sizehints);

    wm_protocols = XInternAtom(dpy, "WM_PROTOCOLS", False);
    wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(dpy, toplevel, &wm_delete_window, 1);

    gc = XCreateGC(dpy, toplevel, 0, NULL);
    XSetForeground(dpy, gc, GetColor("black", BlackPixel(dpy, screen)));

    CardInit();
    StackInit();

    int y = GAMEHEIGHT - GetTextHeight() - 10;
    int width, width2, width3;
    Window undobutton = CreateButtonWindow("Undo (U)", 10, y,
					   SouthWestGravity, &width);
    Window redobutton = CreateButtonWindow("Redo (R)", 10 + width + 5, y,
					   SouthWestGravity, &width2);
    Window restartbutton = CreateButtonWindow("Restart", 
	    				   10 + width + 5 + width2 + 5, y,
				           SouthEastGravity, NULL);
    Window quitbutton = CreateButtonWindow("Quit", 0, 0,
					   SouthEastGravity, &width);
    Window newgamebutton = CreateButtonWindow("New Game", 0, 0,
					      SouthEastGravity, &width2);
    Window autobutton = CreateButtonWindow("Autoplay", 0, 0,
					   SouthWestGravity, &width3);
    XMoveWindow(dpy, quitbutton, GAMEWIDTH - width - 10, y);
    XMoveWindow(dpy, newgamebutton, GAMEWIDTH - width - width2 - 20, y);
    XMoveWindow(dpy, autobutton,
		GAMEWIDTH - width - width2 - width3 - 30, y);

    XMapWindow(dpy, toplevel);
    score = new ScoreRec();

    for (;;) {
	XEvent event;
	GetInterestingEvent(&event);
	if (!CardHandleEvent(&event)) {
	    if (event.type == KeyPress) {
		char buf[100];
		int length;
		length = XLookupString((XKeyEvent *)&event, buf, 100, NULL,
				       NULL);
		for (i=0 ; i<length ; i++) {
		    switch(buf[i]) {
		      case 'u':
		      case 'U':
		      case 'z':
		      case 'Z':
			DoUndo(Fast);
			break;
		      case 'r':
		      case 'R':
			DoRedo(Fast);
			break;
		      case '\016': // ^N
			NewGame();
			break;
		      case '\003': // ^C
		      case '\033': //esc
		      case 'Q':
		      case 'q':
			Quit();
			break;
		    }
		}
	    } else if (event.type == ClientMessage
			&& event.xclient.message_type == wm_protocols
			&& event.xclient.format == 32
			&& *(Atom *)&event.xclient.data == wm_delete_window) {
		Quit();
	    } else if (event.type == ButtonPress) {
		Window w = event.xbutton.window;
		if (w == undobutton) DoUndo(Fast);
		else if (w == redobutton) DoRedo(Fast);
		else if (w == restartbutton) DoRestart();
		else if (w == autobutton) DoAutoPlay();
		else if (w == newgamebutton) NewGame();
		else if (w == quitbutton) Quit();
	    }
	}
    }
}
