/*
 * Author:  Terry Weissman
 *          weissman@sgi.com
 */



#include "seahaven.h"
#include "auto.h"

#include <time.h>		// For time()
#include <sys/time.h>		// For gettimeofday()
#include <math.h>		// For random()
#include <stdlib.h>
#include <unistd.h>
#include <X11/cursorfont.h>

Stack playstack[NUMPLAYSTACK];
Stack singlestack[NUMSINGLESTACK];
Stack donestack[NUMSUITS];


static Stack allstacks[100];
static int numstacks = 0;

static int initializing = false;
static int undoing = false;

class StackPlayRec : public StackRec {
  public:
    StackPlayRec(int x, int y);
    virtual bool addCard(Card, AnimType animate);
    virtual StackType getType();
  private:
};


class StackSingleRec : public StackRec {
  public:
    StackSingleRec(int x, int y);
    virtual bool addCard(Card, AnimType animate);
    virtual StackType getType();
  private:
};


class StackDoneRec : public StackRec {
  public:
    StackDoneRec(int x, int y);
    virtual bool addCard(Card, AnimType animate);
    virtual int getStackingHeight();
    virtual StackType getType();
  private:
};


UndoListRec::UndoListRec() {
    max = 100;
    cards = (Card *) malloc(max * sizeof(Card));
    stacks = (Stack *) malloc(max * sizeof(Stack));
    clear();
}


void UndoListRec::clear() {
    num = 0;
}

void UndoListRec::add(Card c, Stack s) {
    if (num >= max) {
	max += 100;
	cards = (Card *) realloc((char *) cards, max * sizeof(Card));
	stacks = (Stack *) realloc((char *) stacks, max * sizeof(Stack));
    }
    cards[num] = c;
    stacks[num] = s;
    num++;
}

void UndoListRec::addBoundary() {
    if (num > 0 && cards[num - 1] != NULL) add(NULL, NULL);
}


void UndoListRec::doUndo(AnimType animate) {
    initializing = true;	// Hack to fool StackPlayRec::addCard.
    if (num > 0 && cards[num - 1] == NULL) num--;
    while (num > 0 && cards[num - 1] != NULL) {
	num--;
	stacks[num]->addCard(cards[num], animate);
    }
    initializing = false;
    for (int i=0 ; i<NUMPLAYSTACK ; i++) playstack[i]->repositionAll();
}

bool UndoListRec::isEmpty() {
    for (int i = 0 ; i<num ; i++) if (cards[i]) return false;
    return true;
}

UndoListRec undostack;
UndoListRec redostack;

void AddUndo(Card c, Stack s) {
    UndoList list = undoing ? &redostack : &undostack;
    list->add(c, s);
}


void UndoBoundary() {
    undostack.addBoundary();
    redostack.clear();
}


void DoUndo(AnimType animate) {
    undoing = true;
    undostack.doUndo(animate);
    undoing = false;
    redostack.addBoundary();
}

void DoRedo(AnimType animate) {
    redostack.doUndo(animate);
    undostack.addBoundary();
}

void DoRestart() {
    while (!undostack.isEmpty()) DoUndo(Instant);
}

StackRec::StackRec(int newx, int newy) {
    x = newx;
    y = newy;
    numcards = 0;
    allstacks[numstacks++] = this;
}

void StackRec::clear() {
    numcards = 0;
}

int StackRec::getX() {
    return x;
}

int StackRec::getY() {
    return y;
}

Card StackRec::getTopCard() {
    if (numcards == 0) return NULL;
    return cards[numcards - 1];
}

Card StackRec::getCardAbove(Card c) {
    int i;
    for (i=0 ; i<numcards - 1 ; i++) {
	if (cards[i] == c) return cards[i + 1];
    }
    return NULL;
}


void StackRec::repositionAll() {
    for (int i=0 ; i<numcards ; i++) {
	cards[i]->setLoc(x, y + i * getStackingHeight());
    }
}

Card StackRec::popCard() {
    if (numcards == 0) return NULL;
    numcards--;
    return cards[numcards];
}
#include <stdio.h>
bool StackRec::addCard(Card c, AnimType animate) {
    if (c->getStack() == this) Punt("Error in StackRec::addCard!");
    if (animate == Slow || animate == Fast) {
	c->raise();
	int oldx = c->getX();
	int oldy = c->getY();
	int newx = x;
	int newy = y + numcards * getStackingHeight();
        double dist = hypot(oldx-newx, oldy-newy);
	float duration = sqrt(dist)/speedup;
        if (animate == Fast) duration /= 1.5;
	struct timeval t0; gettimeofday(&t0, 0);
	for (;;) {
	    XSync(dpy, 0);
	    usleep(1000);
	    struct timeval t1; gettimeofday(&t1, 0);
	    float t =
	      (t1.tv_sec - t0.tv_sec + (t1.tv_usec - t0.tv_usec)/1000000.0)
	      / duration;
	    if (t >= 1) break;
	    float f = t*t*(3 - t*2);
	    c->setLoc(oldx + int((newx - oldx) * f),
		      oldy + int((newy - oldy) * f));
	}
    }
    if (animate != DontMove) {
	if (numcards > 0) c->raiseAbove(cards[numcards - 1]);
	c->setLoc(x, y + numcards * getStackingHeight());
        if (animate == Slow || animate == Fast) XFlush(dpy);
    }
    Stack old = c->getStack();
    if (old && old->popCard() != c) {
	Punt("Couldn't find card in old stack in StackRec::addCard!");
    }
    AddUndo(c, c->getStack());
    c->setStack(this);
    cards[numcards++] = c;
    return true;
}


int StackRec::getStackingHeight() {
  //    return 22;
  return 28;
}

StackType StackRec::getType() {
    Punt("Invalid call to base class: StackRec::getType()");
    return Single;
}

StackPlayRec::StackPlayRec(int x, int y) : StackRec(x, y) {
}


bool StackPlayRec::addCard(Card c, AnimType animate) {
    if (!initializing) {
	if (numcards == 0) {
	    if (c->getValue() != NUMVALUES - 1) return false;
	} else {
	    if (c->getSuit() != cards[numcards - 1]->getSuit() ||
		   c->getValue() != cards[numcards - 1]->getValue() - 1)
		return false;
	}
    }
    return StackRec::addCard(c, animate);
}

StackType StackPlayRec::getType() {
    return Play;
}


StackSingleRec::StackSingleRec(int x, int y) : StackRec(x, y) {
    XSetWindowAttributes attributes;
    long valuemask = CWBackPixel | CWBorderPixel;
    attributes.background_pixel = GetColor("darkgreen",
					   WhitePixel(dpy, screen));
    attributes.border_pixel = GetColor("green3", BlackPixel(dpy, screen));
    Window w = XCreateWindow(dpy, toplevel, x - 1, y - 1,
			     CARDWIDTH, CARDHEIGHT, 1, (int) CopyFromParent,
			     InputOutput, (Visual *) CopyFromParent,
			     valuemask, &attributes);
    XLowerWindow(dpy, w);
    XMapWindow(dpy, w);
}

bool StackSingleRec::addCard(Card c, AnimType animate) {
    if (numcards) return false;
    return StackRec::addCard(c, animate);
}

StackType StackSingleRec::getType() {
    return Single;
}


StackDoneRec::StackDoneRec(int x, int y) : StackRec(x, y) {
}

bool StackDoneRec::addCard(Card c, AnimType animate) {
    if (numcards == 0) {
	if (c->getValue() != 0) return false;
    } else {
	if (c->getSuit() != cards[numcards - 1]->getSuit() ||
	      c->getValue() != cards[numcards - 1]->getValue() + 1)
	    return false;
    }
    return StackRec::addCard(c, animate);
}

int StackDoneRec::getStackingHeight() {
    return 0;
}

StackType StackDoneRec::getType() {
    return Done;
}


static int Random(int i) {
    return random() % i;
}


void AutoMoves(AnimType animate) {
    bool found;
    int i;
    do {
	found = false;
	for (i=0 ; i<numstacks ; i++) {
	    Card card = allstacks[i]->getTopCard();
	    if (card) {
		Card dcard = donestack[card->getSuit()]->getTopCard();
		if (card->getValue() == (dcard ? dcard->getValue() + 1 : 0)) {
		    donestack[card->getSuit()]->addCard(card, animate);
		    found = true;
		}
	    }
	}
    } while (found);
    for (i=0 ; i<NUMSUITS ; i++) {
	Card card = donestack[i]->getTopCard();
	if (card == NULL ||
	    donestack[i]->getTopCard()->getValue() < NUMVALUES - 1) return;
    }
    score->wonGame();
}


int NumAvailableSingles() {
    int result = 0;
    for (int i=0 ; i<NUMSINGLESTACK ; i++) {
	if (singlestack[i]->getTopCard() == NULL) result++;
    }
    return result;
}


Stack GetAvailableSingle() {
    for (int i=0 ; i<NUMSINGLESTACK ; i++) {
	if (singlestack[i]->getTopCard() == NULL) return singlestack[i];
    }
    return NULL;
}


Stack StackFromPoint(int x, int y) {
    Stack *checklist;
    int num;
    if (y < playstack[0]->getY() - 20) {
	checklist = singlestack;
	num = NUMSINGLESTACK;
    } else {
	checklist = playstack;
	num = NUMPLAYSTACK;
    }
    for (int i=0 ; i<num ; i++) {
	if (x >= checklist[i]->getX() - CARDWIDTH / 2 &&
	    x < checklist[i]->getX() + CARDWIDTH / 2) return checklist[i];
    }
    return NULL;
}


void NewGame(int firstgame) {
    int i, j, k, w;
    if (!firstgame && score) {
	score->lostGame();	// Has no effect if wonGame() already called.
	score->newGame();
    }
    initializing = true;
    for (i=0 ; i<numstacks ; i++) {
	allstacks[i]->clear();
    }
    Card deck[NUMSUITS * NUMVALUES];
    i = 0;
    for (int s = 0 ; s < NUMSUITS ; s++) {
	for (int v = 0 ; v < NUMVALUES ; v++) {
	    deck[i++] = cards[s][v];
	    cards[s][v]->setStack(NULL);
	}
    }
    if (score) score->saveGameBegin();
    for (k=0 ; k<CARDSPERPLAYSTACK ; k++) {
	for (j=0 ; j<NUMPLAYSTACK ; j++) {
	    w = Random(i);
	    playstack[j]->addCard(deck[w]);
	    if (score) score->saveGameCard(deck[w]);
	    deck[w] = deck[--i];
	}
    }
    for (j=0 ; j<i ; j++) {
	singlestack[j + 1]->addCard(deck[j]);
	if (score) score->saveGameCard(deck[j]);
    }
    if (score) score->saveGameEnd();
    initializing = false;

    AutoMoves(Slow);
    undostack.clear();
    redostack.clear();
}



void LoadStacks(Card deck[52]) {
    int i, j, k, w;
    initializing = true;
    for (i=0 ; i<numstacks ; i++) {
	allstacks[i]->clear();
    }
    initializing = true;
    for (int s = 0 ; s < NUMSUITS ; s++) {
	for (int v = 0 ; v < NUMVALUES ; v++) {
	    cards[s][v]->setStack(NULL);
	}
    }
    w = 0;
    for (k=0 ; k<CARDSPERPLAYSTACK ; k++) {
	for (j=0 ; j<NUMPLAYSTACK ; j++) {
	    playstack[j]->addCard(deck[w++]);
	}
    }
    for (j=0 ; j<2 ; j++) {
	singlestack[j + 1]->addCard(deck[w++]);
    }
    initializing = false;

    AutoMoves(Slow);
    undostack.clear();
    redostack.clear();
}


// Force the current win/loss count to save no matter how the program
// exits by doing it in this static destructor:

class StackNoticeExitRec {
  public:
    StackNoticeExitRec() {}
    ~StackNoticeExitRec();
};

StackNoticeExitRec::~StackNoticeExitRec() {
    if (score && score->getGameWonOrLost()) NewGame();
}

StackNoticeExitRec noticeexit;

void StackInit() {
    int i;
    srandom(int(time(NULL)));

    for (i=0 ; i<NUMPLAYSTACK ; i++) {
	playstack[i] = new StackPlayRec(i * (CARDWIDTH+HGAP) + HGAP, 2*VGAP+CARDHEIGHT);
    }
    for (i=0 ; i<NUMSINGLESTACK ; i++) {
	singlestack[i] = new StackSingleRec((i+(NUMPLAYSTACK/2-2)) * (CARDWIDTH+HGAP) + HGAP, 10);
    }
    donestack[0] = new StackDoneRec(0*(CARDWIDTH+HGAP)+HGAP, VGAP);
    donestack[1] = new StackDoneRec(1*(CARDWIDTH+HGAP)+HGAP, VGAP);
    donestack[2] = new StackDoneRec((NUMPLAYSTACK-2)*(CARDWIDTH+HGAP)+HGAP, VGAP);
    donestack[3] = new StackDoneRec((NUMPLAYSTACK-1)*(CARDWIDTH+HGAP)+HGAP, VGAP);
}




void DoAutoPlay(AnimType animate) {
    static Cursor waitcursor = 0;
    if (!waitcursor) waitcursor = XCreateFontCursor(dpy, XC_watch);
    XDefineCursor(dpy, toplevel, waitcursor);
    DoRestart();
    XFlush(dpy);
    int solvable = AutoPlay();
    if (solvable) {
        score->lostGame();
	score->setMessage("Solvable.");
	while (solutionhead) {
	    SolutionLog temp = solutionhead;
	    solutionhead = solutionhead->next;
	    Card card = cards[temp->suit][temp->rank];
	    if (temp->dest == -999) {
		AutoMoves(animate);
		UndoBoundary();
	    } else if (temp->dest == -99) {
		GetAvailableSingle()->addCard(card, animate);
	    } else if (temp->dest <= NUMPLAYSTACK) {
		playstack[temp->dest - 1]->addCard(card, animate);
	    } else {
		if (temp->rank == NUMVALUES - 1) { // King to empty stack.
		    for (int i=0 ; i<NUMPLAYSTACK ; i++) {
			if (playstack[i]->getTopCard() == NULL) {
			    playstack[i]->addCard(card, animate);
			    break;
			}
		    }
		} else {
		    cards[temp->suit][temp->rank+1]->getStack()->addCard(card,animate);
		}
	    }
	    delete temp;
	    XFlush(dpy);	// For debugging.
	}
    } else {
        score->wonGame();
	score->setMessage("Unsolvable.");
    }
    XUndefineCursor(dpy, toplevel);
}
