/*
 * Author:  Terry Weissman
 *          weissman@sgi.com
 */

#include "seahaven.h"

#include <string.h>



unsigned long GetColor(const char *name, unsigned long def) {
    XColor col1, col2;
    if (hascolor && XAllocNamedColor(dpy, cmap, name, &col1, &col2)) {
	return col2.pixel;
    } else {
	return def;
    }
}


void GetInterestingEvent(XEvent *event) {
    XNextEvent(dpy, event);
    while (event->type == Expose) {
	if (event->xexpose.count == 0) score->repaint();
	XNextEvent(dpy, event);
    }
}
	

Window CreateButtonWindow(const char *label, int x, int y, int gravity, int *width) {
    static GC cleargc = NULL;
    int w = GetTextWidth(label) + 4;
    int h = GetTextHeight() + 4;
    if (width) *width = w;
    Pixmap pixmap = XCreatePixmap(dpy, toplevel, w, h,
				  DefaultDepth(dpy, screen));
    if (cleargc == NULL) {
	cleargc = XCreateGC(dpy, toplevel, 0, NULL);
	XSetForeground(dpy, cleargc,  backpixel);
    }
    XFillRectangle(dpy, pixmap, cleargc, 0, 0, w, h);
    DrawText(pixmap, 2, 2 + GetTextAscent(), label);

    XSetWindowAttributes attributes;
    long valuemask = CWEventMask | CWBackPixmap | CWWinGravity;
    attributes.event_mask = ButtonPressMask;
    attributes.background_pixmap = pixmap;
    attributes.win_gravity = gravity;
    Window window = XCreateWindow(dpy, toplevel, x, y,
				  w, h, 1, (int) CopyFromParent,
				  InputOutput, (Visual *) CopyFromParent,
				  valuemask, &attributes);
    XMapWindow(dpy, window);
    return window;
}

#ifdef USE_XFT
#    include <X11/Xft/Xft.h>

typedef XftFont* FONT;

static FONT getFont() {
    static FONT font = 0;
    if (!font) {
	font = XftFontOpen(dpy, screen,
			   XFT_FAMILY, XftTypeString, "sans",
			   XFT_WEIGHT, XftTypeInteger, XFT_WEIGHT_MEDIUM,
			   XFT_SLANT, XftTypeInteger, XFT_SLANT_ROMAN,
			   //XFT_PIXEL_SIZE, XftTypeDouble, 12.0,
			   (void*)0);
    }
    return font;
}

extern int GetTextWidth(const char* text) {
    XGlyphInfo i;
    XftTextExtentsUtf8(dpy, getFont(), (XftChar8*)text, strlen(text), &i);
    return i.xOff;
}

extern int GetTextHeight() {
    FONT font = getFont();
    return font->ascent + font->descent;
}

extern int GetTextAscent() {
    return getFont()->ascent;
}

extern int DrawText(Pixmap pixmap, int x, int y, const char* text) {
    static XftDraw* xftgc;
    static XftColor color;
    if (!xftgc) {
	xftgc = XftDrawCreate(dpy, pixmap, DefaultVisual(dpy, screen), cmap);
	color.pixel = 0;
	color.color.red   = 0;
	color.color.green = 0;
	color.color.blue  = 0;
	color.color.alpha = 0xffff;
    } else {
	XftDrawChange(xftgc, pixmap);
    }

    if (text && text[0])
	XftDrawStringUtf8(xftgc, &color, getFont(), x, y,
			  (XftChar8*)text, strlen(text));
}

#else
    
typedef XFontStruct* FONT;

static FONT getFont() {
    static FONT font = 0;
    if (!font) {
	font = XLoadQueryFont
	    (dpy, "-*-helvetica-medium-r-normal--*-120-*-*-p-*-iso8859-1");
	if (font == NULL) font = XLoadQueryFont(dpy, "fixed");
	if (font == NULL) Punt("Can't find a font!");
	XSetFont(dpy, gc, getFont()->fid);
    }
    return font;
}

extern int GetTextWidth(const char* text) {
    return XTextWidth(getFont(), text, strlen(text));
}

extern int GetTextHeight() {
    FONT font = getFont();
    return font->ascent + font->descent;
}

extern int GetTextAscent() {
    return getFont()->ascent;
}

extern int DrawText(Pixmap pixmap, int x, int y, const char* text)
{
    if (text && text[0])
	XDrawString(dpy, pixmap, gc, x, y, text, strlen(text));
}

#endif
